@@ -6,8 +6,6 @@
 
 #include "irrTypes.h"
 
-namespace irr
-{
 namespace gui
 {
 enum EGUI_ALIGNMENT
@@ -32,4 +30,3 @@ const c8 *const GUIAlignmentNames[] = {
 	};
 
 } // namespace gui
-} // namespace irr
