@@ -6,8 +6,6 @@
 
 #include "ISceneNode.h"
 
-namespace irr
-{
 namespace scene
 {
 
@@ -33,4 +31,3 @@ public:
 };
 
 } // end namespace scene
-} // end namespace irr
