@@ -6,8 +6,6 @@
 
 #include "vector3d.h"
 
-namespace irr
-{
 namespace scene
 {
 
@@ -39,4 +37,3 @@ private:
 };
 
 } // end namespace scene
-} // end namespace irr
