@@ -10,8 +10,6 @@
 
 #include "IImageLoader.h"
 
-namespace irr
-{
 namespace video
 {
 
@@ -31,4 +29,3 @@ public:
 };
 
 } // end namespace video
-} // end namespace irr
