@@ -1,8 +1,9 @@
 #pragma once
 
-#include "irrlichttypes_extrabloated.h"
 #include <algorithm>
-#include <string>
+#include <IGUIElement.h>
+#include <IGUIEnvironment.h>
+
 
 class ISimpleTextureSource;
 
