@@ -2,7 +2,6 @@
 // SPDX-License-Identifier: LGPL-2.1-or-later
 // Copyright (C) 2013 celeron55, Perttu Ahola <celeron55@gmail.com>
 
-#include "irrlichttypes_extrabloated.h"
 #include "lua_api/l_util.h"
 #include "lua_api/l_internal.h"
 #include "lua_api/l_settings.h"
@@ -27,8 +26,7 @@
 #include "config.h"
 #include "version.h"
 #include "util/hex.h"
-#include "util/sha1.h"
-#include "my_sha256.h"
+#include "util/hashing.h"
 #include "util/png.h"
 #include "player.h"
 #include "daynightratio.h"
@@ -56,7 +54,13 @@ int ModApiUtil::l_log(lua_State *L)
 		auto name = readParam<std::string_view>(L, 1);
 		text = readParam<std::string_view>(L, 2);
 		if (name == "deprecated") {
-			log_deprecated(L, text, 2);
+			// core.log("deprecated", message [, stack_level])
+			// Level 1 - immediate caller of core.log (probably engine code);
+			// Level 2 - caller of the function that called core.log, and so on
+			int stack_level = readParam<int>(L, 3, 2);
+			if (stack_level < 1)
+				throw LuaError("invalid stack level");
+			log_deprecated(L, text, stack_level);
 			return 0;
 		}
 		level = Logger::stringToLevel(name);
@@ -494,18 +498,13 @@ int ModApiUtil::l_request_insecure_environment(lua_State *L)
 {
 	NO_MAP_LOCK_REQUIRED;
 
-	// Just return _G if security is disabled
-	if (!ScriptApiSecurity::isSecure(L)) {
-		lua_getglobal(L, "_G");
-		return 1;
-	}
-
-	if (!ScriptApiSecurity::checkWhitelisted(L, "secure.trusted_mods")) {
-		return 0;
+	if (ScriptApiSecurity::isSecure(L)) {
+		if (!ScriptApiSecurity::checkWhitelisted(L, "secure.trusted_mods"))
+			return 0;
 	}
 
 	// Push insecure environment
-	lua_rawgeti(L, LUA_REGISTRYINDEX, CUSTOM_RIDX_GLOBALS_BACKUP);
+	ScriptApiSecurity::getGlobalsBackup(L);
 	return 1;
 }
 
@@ -545,12 +544,7 @@ int ModApiUtil::l_sha1(lua_State *L)
 	bool hex = !lua_isboolean(L, 2) || !readParam<bool>(L, 2);
 
 	// Compute actual checksum of data
-	std::string data_sha1;
-	{
-		SHA1 ctx;
-		ctx.addBytes(data);
-		data_sha1 = ctx.getDigest();
-	}
+	std::string data_sha1 = hashing::sha1(data);
 
 	if (hex) {
 		std::string sha1_hex = hex_encode(data_sha1);
@@ -569,10 +563,7 @@ int ModApiUtil::l_sha256(lua_State *L)
 	auto data = readParam<std::string_view>(L, 1);
 	bool hex = !lua_isboolean(L, 2) || !readParam<bool>(L, 2);
 
-	std::string data_sha256;
-	data_sha256.resize(SHA256_DIGEST_LENGTH);
-	SHA256(reinterpret_cast<const unsigned char*>(data.data()), data.size(),
-		reinterpret_cast<unsigned char *>(data_sha256.data()));
+	std::string data_sha256 = hashing::sha256(data);
 
 	if (hex) {
 		lua_pushstring(L, hex_encode(data_sha256).c_str());
